#include <stdio.h>

#define LONG long long int
#define ULONG unsigned long long int

unsigned LONG seed;
unsigned LONG mult;

void vecprint64s(LONG* A, int n);

#define min64s gbmin642
#define max64s gbmax642
inline LONG min64s( LONG a, LONG b ) { if( a<b ) return a; else return b; }
inline LONG max64s( LONG a, LONG b ) { if( a>b ) return a; else return b; }

#define min gbmin
#define max gbmax
inline int max(int a, int b) { if(a>b) return a; else return b; }
inline int min(int a, int b) { if(a<b) return a; else return b; }
#define add64s gbadd64s
#define sub64s gbsub64s
inline LONG add64s(LONG a, LONG b, LONG p) { LONG t; t = (a-p)+b; t += (t>>63) & p; return t; }
inline LONG sub64s(LONG a, LONG b, LONG p) { LONG t = a-b; t += (t>>63) & p; return t; }
//inline LONG mul64s(LONG a, LONG b, LONG p) { LONG t; t = (__int128_t) a * b % p; return t; }
#define mul64s gbmul64s
inline LONG mul64s(LONG a, LONG b, LONG p)
{
         LONG q, r;
         __asm__ __volatile__(           \
         "       mulq    %%rdx           \n\t" \
         "       divq    %4              \n\t" \
         : "=a"(q), "=d"(r) : "0"(a), "1"(b), "rm"(p));
         return r;
}

#define M_INT long long int

typedef struct P_INT P_INT;
struct P_INT {
 	M_INT p;	/* prime */
 	M_INT s;	/* shift */
 	M_INT d;	/* normalized */
 	M_INT v;	/* reciprocal */
};

/* initialize mod p struct */
#define modinit64s gbmodinit64s
inline P_INT modinit64s(M_INT p)
{
        P_INT m;

        /* most significant bit location 63..0 */
        __asm__ __volatile__("bsrq %0, %0" : "=r"(m.s) : "0"(p));
        m.p = p;
        m.s = 64-m.s-1;
        m.d = p << m.s;

        /* reciprocal: (2^128-1)/n - 2^64 */
        /* [(2^64-n-1)*2^64 + (2^64-1)]/n */
        __asm__ __volatile__(\
        "       movq    %1, %%rdx       \n\t" \
        "       addq    $1, %%rdx       \n\t" \
        "       negq    %%rdx           \n\t" \
        "       movq    $-1, %%rax      \n\t" \
        "       divq    %1              \n\t" \
                : "=a"(m.v) : "r"(m.d) : "%rdx");
        return m;
}

#define mulmodinv64s gbmulmodinv64s
inline M_INT mulmodinv64s(M_INT a, M_INT b, P_INT m)
{
        M_INT q, r;
        /* 0: compute  a*b = u1:u0, lshift by s, r9:r8 = (u1+1):u0 */
        /* 1: compute u1*v = q1:q0, q1:q0 += (u1+1):u0, in rdx:rax */
        /* 2: compute    r = u0-q1*d in r8 */
        /* 3: if r > q0 then q1 = q1-1, r = r+d (possible) */
        /* 4: if r >= d then q1 = q1+1, r = r-d (unlikely) */
        __asm__ __volatile__(\
        "0:     mulq    %%rdx                   \n\t" \
        "       shldq   %%cl,  %%rax, %%rdx     \n\t" \
        "       shlq    %%cl,  %%rax            \n\t" \
        "       leaq    1(%%rdx), %%r9          \n\t" \
        "       movq    %%rax, %%r8             \n\t" \
        "1:     movq    %%rdx, %%rax            \n\t" \
        "       mulq    %5                      \n\t" \
        "       addq    %%r8,  %%rax            \n\t" \
        "       adcq    %%r9,  %%rdx            \n\t" \
        "2:     imulq   %4,    %%rdx            \n\t" \
        "       subq    %%rdx, %%r8             \n\t" \
        "3:     cmpq    %%rax, %%r8             \n\t" \
        "       jbe     4f                      \n\t" \
        "       addq    %4,    %%r8             \n\t" \
        "       decq    %%rax                   \n\t" \
        "4:     cmpq    %4,    %%r8             \n\t" \
        "       jae     5f                      \n\t" \
        "       jmp     6f                      \n\t" \
        "5:     subq    %4,    %%r8             \n\t" \
        "       incq    %%rax                   \n\t" \
        "6:     shrq    %%cl,  %%r8             \n\t" \
        "       movq    %%r8,  %%rdx            \n\t" \
        : "=a"(q), "=d"(r)
        :  "0"(a),  "1"(b), "r"(m.d), "r"(m.v), "c"(m.s)
        : "%r8", "%r9", "%r10");
        /* q = quotient */
        return r;
}

        /* z += a1:a0 */
        #define zadd(z,a0,a1) __asm__(\
        "       addq    %4, %0  \n\t" \
        "       adcq    %5, %1  \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]) : "0"(z[0]), "1"(z[1]), "r"(a0), "r"(a1))

        /* z -= a1:a0 */
        #define zsub(z,a0,a1) __asm__(\
        "       subq    %4, %0  \n\t" \
        "       sbbq    %5, %1  \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]) : "0"(z[0]), "1"(z[1]), "r"(a0), "r"(a1))

        /* z = a*b */
        #define zmul(z,a,b) __asm__(\
        "       mulq    %%rdx   \n\t" \
                : "=a"(z[0]), "=d"(z[1]) : "a"(a), "d"(b))

        /* z += a*b */
        #define zfma(z,a,b) do {        \
        unsigned long u,v;              \
        __asm__(                        \
        "       mulq    %%rdx           \n\t" \
        "       addq    %%rax, %0       \n\t" \
        "       adcq    %%rdx, %1       \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]), "=a"(u), "=d"(v) : "0"(z[0]), "1"(z[1]), "a"(a), "d"(b));\
        } while (0)

        /* z -= a*b */
        #define zfms(z,a,b) do {        \
        unsigned long u,v;              \
        __asm__(                        \
        "       mulq    %%rdx           \n\t" \
        "       subq    %%rax, %0       \n\t" \
        "       sbbq    %%rdx, %1       \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]), "=a"(u), "=d"(v) : "0"(z[0]), "1"(z[1]), "a"(a), "d"(b));\
        } while (0)

        /* z[0] = z % p */
        /* z[1] = z / p */
        /* quotient can overflow */
        #define zdiv(z,p) __asm__(\
        "       divq    %4      \n\t" \
                : "=a"(z[1]), "=d"(z[0]) : "a"(z[0]), "d"(z[1]), "r"(p))

        /* z = z % p safe */
        #define zmod(z,p) __asm__(\
        "       divq    %4      \n\t" \
        "       xorq    %0, %0  \n\t" \
                : "=a"(z[1]), "=d"(z[0]) : "a"(z[0]), "d"(z[1] < p ? z[1] : z[1] % p), "r"(p))

        /* z = z << s */
        #define zshl(z,s) __asm__(\
        "       shldq   %%cl, %0, %1    \n\t" \
        "       shlq    %%cl, %0        \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]) : "0"(z[0]), "1"(z[1]), "c"(s))

LONG neg64sd(LONG a, LONG p) { return (a==0) ? 0 : p-a; }

LONG rand64s(LONG p);

void randpoly64s(int n, int* DEG, int t, LONG* A, LONG* X, LONG p);
void polcopy64s(LONG* A, int d, LONG* B);
void veccopy64s(LONG *A, int n, LONG *B);
int vec2partition64s(LONG* D, LONG* C, int n);
void sort2vec64s(LONG* D, LONG* C, int n);
int pol2simp64s(LONG* C, LONG* X, int n, LONG p);
int polsparsemul64s(LONG* A, LONG* X, int m, LONG* B, LONG* Y, int n, LONG p, LONG* C, LONG* Z);
int polnsparsedeg64s(LONG* X, int t, int n, int xi);

/*----------needed for gcd()----------*/

LONG modinv64s(LONG c, LONG p);
void monic64s(LONG* A, int d, LONG p);

int pol2eval64s(
    LONG* A, LONG* X, int na,
    LONG* f, int d,
    LONG* alpha, int numbits,
    LONG p)
{
    // X = [ (i1,j1) | (i2,j2) | ... ]
    int i,j,dx,dy;
    LONG mask,x;
    __int128_t t, M;

    for( i=0; i<=d; i++ )
        f[i] = 0;

    mask = (1 << numbits) - 1;

    M = p; M = M << 64;
    for( i=0; i<na; )
    {
        x = X[i];
        dy = x & mask;
        dx = x >> numbits;
        t = M - (__int128_t) alpha[dy]*A[i];

        for( i++; i<na && (X[i]>>numbits) == dx; i++ )
        {
            dy = X[i] & mask;
            t = t - (__int128_t) alpha[dy]*A[i];
            t += (t>>127) & M;
        }

        x = t % p;
        x = neg64sd(x,p);
        f[dx] = add64s(x,f[dx],p); // add in case terms not sorted in lex order
    }

    for( ; d>=0 && f[d]==0; d-- );
    return d;
}

int polsubmulP( LONG *A, LONG *B, LONG a, LONG b, int dA, int dB, LONG p, P_INT P );
int poldiv64s(LONG* A, LONG* B, int da, int db, LONG p);
int polgcd64s( LONG * A, LONG * B, int da, int db, LONG p );

LONG dotprod64s(LONG* u, LONG* v, int n, LONG p)
{
    int i;
    LONG x;
    __int128_t t, M;

    if( n==0 ) return 0;
    M = p; M = M << 64;
    for( t=0,i=0; i<n; i++ )
    {
        t = t - (__int128_t) u[i]*v[i];
        t += (t>>127) & M;
    }
    x = t % p;
    x = neg64sd(x,p);
    return x;
}

void polscamul64s(LONG x, LONG* A, int d, LONG p);

void makem64s(LONG* alpha, int n, LONG* M, LONG p)
{
    // M[i] = product( alpha[n-1]-alpha[i], i=0..n-2 ) mod p
    int i;
    M[0] = 1;
    for( i=0; i<n-1; i++ )
        M[i+1] = mul64s(sub64s(alpha[n-1],alpha[i],p),M[i],p);
    return;
}

void pol2int64s(LONG** V, int N, LONG* y, LONG* m, LONG inv, int n, LONG p)
{   // incrementally interpolate images
    int i;
    LONG* v;
    for( i=0; i<N; i++ )
    {
        v = V[i];
        if( n==1 ) v[0] = y[i];
        else v[n-1] = mul64s(inv, sub64s(y[i],dotprod64s(v,m,n-1,p),p), p);
    }
    return;
}

void stdbasis64s(LONG* alpha, LONG* m, int n, LONG** B, LONG p)
{
    int i,j;
    LONG a;
    // expand( product( (x-alpha[j-1]), j=1..n-1 ) )
    for( i=0; i<n; i++ )
    {
        if( i==0 )
        {
            m[0] = 1;
            B[0][0] = 1;
            continue;
        }

        for( j=i; j>0; j-- )
            m[j] = m[j-1];

        m[0] = 0;
        a = alpha[i-1];
        for( j=0; j<i; j++ )
            m[j] = sub64s(m[j],mul64s(a,m[j+1],p),p);
        for( j=0; j<=i; j++ )
            B[j][i-j] = m[j]; // transpose
    }
    return;
}

void basismul64s(LONG* v, int n, LONG** B, LONG p)
{   int i;
    // B[i] has length n-i
    for( i=0; i<n; i++ ) v[i] = dotprod64s(v+i,B[i],n-i,p);
    return;
}

// V is the bivariate polynomial represented by array of pointers where v[i] contains the terms with degree i in x
// D contains the degrees of the v[i] so that D[i] = degree of coeff(V, x^i) in y
int getcontent64s( LONG **V, LONG *D, int dx, LONG *content, LONG *temp, LONG p )
{
    int i, dcont;
    LONG *v, d;

    for( i=dx; i>=0; i-- )
    {
        v = V[i];
        d = D[i];
        if (i == dx)
        {
            polcopy64s(v, d, content);
            dcont = d;
        }
        else if (d >= 0)
        {
            if (dcont == 0) return dcont;
            polcopy64s(v, d, temp);
            dcont = polgcd64s(content, temp, dcont, d, p);
        }
    }

    return dcont;
}

// see above, this function assumes content, dcont were computed by getcontent64s(..)
void primpart64s( LONG **V, LONG *D, int dx, LONG *content, int dcont, LONG p )
{
    int i;
    LONG *v, d;

    if (dcont <= 0) return;

    for( i=dx; i>=0; i-- )
    {
        v = V[i];
        d = D[i];
        poldiv64s(v, content, d, dcont, p);
        veccopy64s(v+dcont, d+1-dcont, v);
        D[i] -= dcont;
    }
}

/*----------needed for gcd()----------*/

int gcdbiv(
    LONG* A, LONG* X, int na, int dax, int day, // sparse in Zp[x,y]
    LONG* B, LONG* Y, int nb, int dbx, int dby, // sparse in Zp[x,y]
    LONG* C, int dcy, // dense in Zp[y]
    LONG** pGC, LONG** pGX, // pointers to answer
    LONG* space, LONG** ptrSpace,  // space for output in Zp[x,y]
    int dgx, int dgy,
    LONG alpha, int numbits,
    LONG p,
    int info)
{
    int N, dy, d, dg, ng;
    LONG mask;
    LONG *P, *m, *a, *M, *f, *g, *D, *W, *BasisSpace, *NewtonSpace, *v;
    LONG *G, *Z;
    LONG **V, **Basis;
    int i,j,k,count;
    LONG beta,gamma,inv;
    LONG *content, *temp;
    int dcont;

    if( info ) printf("gcd called  dgx=%d  dgy=%d\n",dgx,dgy);
    if( info ) printf("dax=%d, day=%d\n",dax,day);
    if( info ) printf("dbx=%d, dby=%d\n",dbx,dby);

    monic64s(C, dcy, p);

    mask = (1 << numbits) - 1;

    N = 1 + min64s( min64s(day,dby), dgy+dcy );
    dy = max64s(day,dby);

    P = space;
    m = P+dy+1;
    a = m+N+1;
    M = a+N+1;
    f = M+N+1;
    g = f+dax+1;
    D = g+dbx+1;
    NewtonSpace = D+dgx+1;
    BasisSpace = NewtonSpace+(dgx+1)*(N+1);
    G = BasisSpace+((dy+1)*(dy+2))/2;
    Z = G+(N+1)*(dgx+1);
    content = Z+(N+1)*(dgx+1);
    temp = content+dy+1;

    V = ptrSpace;
    W = NewtonSpace;
    for( i=0; i<=dgx; i++ )
    {
        V[i] = W; W = W+(N+1); // array(N+1);
    }

    i = 0; count = 0;
    beta = 1;
    while( i < N )
    {
        beta = mul64s(alpha,beta,p);
        a[i] = beta;

        for( P[0]=1,j=1; j<=dy; j++ )
            P[j] = mul64s(beta,P[j-1],p);

        d = pol2eval64s(A,X,na,f,dax,P,numbits,p);  if( d<dax ) continue;
        d = pol2eval64s(B,Y,nb,g,dbx,P,numbits,p);  if( d<dbx ) continue;
        dg = polgcd64s(g,f,dbx,dax,p);

        if( dg>dgx ) continue;
        if( dg<dgx ) { dgx = dg; i = 0; continue; }

        i++;

        gamma = dotprod64s(C,P,dcy+1,p);   // gamma = C(beta)
        polscamul64s(gamma,g,dg,p);
        makem64s(a, i, m, p);              //printf("m = "); vecprint32s(m,i);
        inv = modinv64s(m[i-1],p);
        pol2int64s(V,dg+1,g,m,inv,i,p);    // incremental interpolation*/
    }

    dy = -1;
    for( i=0; i<=dgx; i++ )
    {
        v = V[i];
        for( d=N-1; d>=0 && v[d]==0; d-- );
            D[i] = d;
        //if( info>1 ) { printf("i=%d: degy=%d  ",i,d); vecprint32s(v,d+1); }
        if( d>dy ) dy = d;
    }
    if( info ) printf("N=%d dy=%d\n",N,dy);
    if( info ) printf("Convert to standard basis\n");

    Basis = V+dgx+1;
    W = BasisSpace;
    for( i=0; i<=dy; i++ ) { Basis[i] = W; W = W+(dy+1-i); } //array(dy+1-i);
    stdbasis64s( a, M, dy+1, Basis, p );

    for( ng=0,i=dgx; i>=0; i-- )
    {
        v = V[i];
        d = D[i];
        basismul64s( v, d+1, Basis, p );
        //if( info>1 ) { printf("x^%2d: ",i); vecprint32s( v, d+1 ); }
        for( j=0; j<=d; j++ ) if( v[j] ) ng++;
    }

    if( info ) printf("Computing content of answer\n");
    dcont = getcontent64s(V, D, dgx, content, temp, p);

    if( info ) printf("Getting primpart by dividing by content\n");
    primpart64s(V, D, dgx, content, dcont, p);

    if( info ) printf("Build answer in sparse format in G,Z\n");
    for( k=0,i=dgx; i>=0; i-- )
    {
        v = V[i];
        for( j=D[i]; j>=0; j-- )
        {
            if( v[j] )
            {
                G[k] = v[j];
                Z[k] = ((LONG)i<<numbits) | j;
                k++;
            }
        }
    }

    *pGC = G;
    *pGX = Z;

    return k;
}

LONG gcdbivspace(int dax, int day, int dbx, int dby, int dgx, int dgy)
{
    int dy,N;
    N = 1 + max64s( min64s(day,dby), dgy );
    dy = max64s(day,dby);
    return dy+1+
        N+1+
        N+1+
        N+1+
        dax+1+
        dbx+1+
        dgx+1+
        (dgx+1)*(N+1)+
        ((dy+1)*(dy+2))/2+
        2*(N+1)*(dgx+1)+
        2*(dy+1);
}
