
#include <stdio.h>
#define LONG long long int
#define ULONG long long int
void Kronecker64s( int *D, ULONG *M, int n )
{
     int i;
     // D = [d1,d2,d3,d4]   M = [1,(d1+1),(d1+1)*(d2+1),2^k>=(d1+1)*(d2+1)*(d3+1)]
     for( M[0]=1, i=1; i<n; i++ ) M[i] = M[i-1]*(D[i-1]+1);
}

void KroneckerMap64s( ULONG *X, int t, ULONG *M, int n )
{
     int i,j,numbits;
     ULONG mask,x,y,d;
     numbits = 63/n;
     mask = (((ULONG) 1) << numbits) - 1 ;
     for( i=0; i<t; i++ ) {
        x = X[i];
        y = 0;
        for( j=0; j<n; j++ ) {
            if( j<n-1 ) d = x & mask; else d = x; // deg[xj](X[i])
            y = y + d*M[j];
            x = x >> numbits;
        }
        X[i] = y;
     }
     return;
}

void KroneckerInv64s( ULONG *Y, int t, ULONG *M, int n )
{    // X = Y is allowed
     int i,j,numbits;
     ULONG x,y,d;
     numbits = 63/n;
     for( i=0; i<t; i++ ) {
        y = Y[i];
        x = 0;
        for( j=n-1; j>0; j-- ) {
            d = y / M[j];
            y = y - d*M[j];
            x = (x << numbits) | d;
        }
        x = (x << numbits) | y;
        Y[i] = x;
     }
     return;
}

/*

int main() {

ULONG X[1], Y[1], M[5], N[5], x, y, mask;
int i,n,k,numbits, D[5];

n = 5;
D[0] = 99;
D[1] = 99;
D[2] = 99;
D[3] = 99;
D[4] = 99;
k = Kronecker64s( D, M, n );
printf("k=%d\n",k);
printf("M="); for( i=0; i<n; i++ ) printf("%lld ",M[i]); printf("\n");
N[0] = 5;
N[1] = 10;
N[2] = 99;
N[3] = 88;
N[4] = 10;
printf("N="); for( i=0; i<n; i++ ) printf("%lld ",N[i]); printf("\n");
numbits = 63/n;
printf("numbits=%d\n",numbits);
x = 0;
for( i=0; i<n; i++ ) x = (x << numbits) | N[4-i];
X[0] = x;
printf("x=%lld\n",x);
KroneckerMap64s( X, Y, 1, M, n );
printf("K(x)=%lld\n",Y[0]);
printf("dx=%lld  should be 10\n", Y[0]>>k );
KroneckerInv64s( Y, Y, 1, M, n );
printf("Krinv(x)=%lld\n",Y[0]);

N[0] = 99;
N[1] = 99;
N[2] = 99;
N[3] = 99;
N[4] = 99;
printf("N="); for( i=0; i<n; i++ ) printf("%lld ",N[i]); printf("\n");
x = 0;
for( i=0; i<n; i++ ) x = (x << numbits) | N[4-i];
X[0] = x;
printf("x=%lld\n",x);
KroneckerMap64s( X, Y, 1, M, n );
printf("K(x)=%lld\n",Y[0]);
KroneckerInv64s( Y, X, 1, M, n );
printf("Kinv(y)=%lld  (x=%lld)\n",X[0],x);
printf("dx=%lld  should be 99\n", Y[0]>>k );
mask = (((ULONG) 1) << k) - 1;
x = Y[0] & mask;
printf("DLOG=%lld\n",x);
Y[0] = x; // from dlogs
KroneckerInv64s( Y, Y, 1, M, n );
printf("Krinv(x)=%lld\n",Y[0]);

return 1;
}

*/
