#define M_INT long long int
#define LONG long long int
#define ULONG unsigned long long int

#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>

LONG * array( int n );
LONG ** arrayofptrs( int n );

LONG rand64s(LONG p) {
    LONG x,y;
    extern ULONG seed, mult;
    seed = mult*seed;
    x = seed >> 32;
    seed = mult*seed;
    y = seed >> 32;
    x = (x<<31) | y;
    x = x % p;
    return(x);
}

/******************************************************************************************/
/*  Roman's assembler utilities                                                           */
/******************************************************************************************/

#define WORDSIZE 64
typedef unsigned long M_UINT;

/* [x0:x1] := a*b */
#define MUL211(x0,x1,a,b) __asm__(\
"       mulq    %3      \n\t"   \
: "=a"(x0), "=d"(x1) : "0"(a), "r"(b) : "cc")

/* x0 := [x0:x1] / a */
/* x1 := [x0:x1] % a */
#define DIV21H(x0,x1,a) __asm__(\
"       divq    %2      \n\t"   \
: "=a"(x0), "=d"(x1) : "r"(a), "0"(x0), "1"(x1) : "cc")

/* floor of log2 */
M_UINT ilog2new(M_UINT x)
{
        M_UINT y;
        for (y=0; x >>= 1; y++);
        return y;
}

/* reciprocal of full word */
M_UINT recipnew(M_UINT p)
{
        M_UINT u0, u1;
        u1 = -(p+1);
        u0 = -1;
        DIV21H(u0,u1,p);
        return u0;
}

/* compute reciprocal and shift */
void init(M_UINT d, M_UINT *r, M_UINT *s)
{
        int l;
        l = ilog2new(d);
        if( d != (M_UINT)1 << l ) l++;
        d = d << (WORDSIZE-l);
        *r = recipnew(d);
        *s = l-1;
}

/* compute remainder via reciprocal */
inline M_UINT rem(M_UINT n, M_UINT d, M_UINT r, M_UINT s)
{
        M_UINT u0,u1,q;
        MUL211(u0,u1,n,r);
        q = (n - u1) >> 1;
        q = (q + u1) >> s;
        return n - q*d;
}

        /* z += a1:a0 */
        #define zadd(z,a0,a1) __asm__(\
        "       addq    %4, %0  \n\t" \
        "       adcq    %5, %1  \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]) : "0"(z[0]), "1"(z[1]), "r"(a0), "r"(a1))

        /* z -= a1:a0 */
        #define zsub(z,a0,a1) __asm__(\
        "       subq    %4, %0  \n\t" \
        "       sbbq    %5, %1  \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]) : "0"(z[0]), "1"(z[1]), "r"(a0), "r"(a1))

        /* z = a*b */
        #define zmul(z,a,b) __asm__(\
        "       mulq    %%rdx   \n\t" \
                : "=a"(z[0]), "=d"(z[1]) : "a"(a), "d"(b))

        /* z += a*b */
        #define zfma(z,a,b) do {        \
        unsigned long u,v;              \
        __asm__(                        \
        "       mulq    %%rdx           \n\t" \
        "       addq    %%rax, %0       \n\t" \
        "       adcq    %%rdx, %1       \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]), "=a"(u), "=d"(v) : "0"(z[0]), "1"(z[1]), "a"(a), "d"(b));\
        } while (0)

        /* z -= a*b */
        #define zfms(z,a,b) do {        \
        unsigned long u,v;              \
        __asm__(                        \
        "       mulq    %%rdx           \n\t" \
        "       subq    %%rax, %0       \n\t" \
        "       sbbq    %%rdx, %1       \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]), "=a"(u), "=d"(v) : "0"(z[0]), "1"(z[1]), "a"(a), "d"(b));\
        } while (0)

        /* z[0] = z % p */
        /* z[1] = z / p */
        /* quotient can overflow */
        #define zdiv(z,p) __asm__(\
        "       divq    %4      \n\t" \
                : "=a"(z[1]), "=d"(z[0]) : "a"(z[0]), "d"(z[1]), "r"(p))

        /* z = z % p safe */
        #define zmod(z,p) __asm__(\
        "       divq    %4      \n\t" \
        "       xorq    %0, %0  \n\t" \
                : "=a"(z[1]), "=d"(z[0]) : "a"(z[0]), "d"(z[1] < p ? z[1] : z[1] % p), "r"(p))

        /* z = z << s */
        #define zshl(z,s) __asm__(\
        "       shldq   %%cl, %0, %1    \n\t" \
        "       shlq    %%cl, %0        \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]) : "0"(z[0]), "1"(z[1]), "c"(s))

typedef struct P_INT P_INT;
struct P_INT {
 	M_INT p;	/* prime */
 	M_INT s;	/* shift */
 	M_INT d;	/* normalized */
 	M_INT v;	/* reciprocal */
};

/* initialize mod p struct */
inline P_INT modinit64s(M_INT p)
{
        P_INT m;

        /* most significant bit location 63..0 */
        __asm__ __volatile__("bsrq %0, %0" : "=r"(m.s) : "0"(p));
        m.p = p;
        m.s = 64-m.s-1;
        m.d = p << m.s;

        /* reciprocal: (2^128-1)/n - 2^64 */
        /* [(2^64-n-1)*2^64 + (2^64-1)]/n */
        __asm__ __volatile__(\
        "       movq    %1, %%rdx       \n\t" \
        "       addq    $1, %%rdx       \n\t" \
        "       negq    %%rdx           \n\t" \
        "       movq    $-1, %%rax      \n\t" \
        "       divq    %1              \n\t" \
                : "=a"(m.v) : "r"(m.d) : "%rdx");
        return m;
}

inline M_INT mulmodinv64s(M_INT a, M_INT b, P_INT m)
{
        M_INT q, r;
        /* 0: compute  a*b = u1:u0, lshift by s, r9:r8 = (u1+1):u0 */
        /* 1: compute u1*v = q1:q0, q1:q0 += (u1+1):u0, in rdx:rax */
        /* 2: compute    r = u0-q1*d in r8 */
        /* 3: if r > q0 then q1 = q1-1, r = r+d (possible) */
        /* 4: if r >= d then q1 = q1+1, r = r-d (unlikely) */
        __asm__ __volatile__(\
        "0:     mulq    %%rdx                   \n\t" \
        "       shldq   %%cl,  %%rax, %%rdx     \n\t" \
        "       shlq    %%cl,  %%rax            \n\t" \
        "       leaq    1(%%rdx), %%r9          \n\t" \
        "       movq    %%rax, %%r8             \n\t" \
        "1:     movq    %%rdx, %%rax            \n\t" \
        "       mulq    %5                      \n\t" \
        "       addq    %%r8,  %%rax            \n\t" \
        "       adcq    %%r9,  %%rdx            \n\t" \
        "2:     imulq   %4,    %%rdx            \n\t" \
        "       subq    %%rdx, %%r8             \n\t" \
        "3:     cmpq    %%rax, %%r8             \n\t" \
        "       jbe     4f                      \n\t" \
        "       addq    %4,    %%r8             \n\t" \
        "       decq    %%rax                   \n\t" \
        "4:     cmpq    %4,    %%r8             \n\t" \
        "       jae     5f                      \n\t" \
        "       jmp     6f                      \n\t" \
        "5:     subq    %4,    %%r8             \n\t" \
        "       incq    %%rax                   \n\t" \
        "6:     shrq    %%cl,  %%r8             \n\t" \
        "       movq    %%r8,  %%rdx            \n\t" \
        : "=a"(q), "=d"(r)
        :  "0"(a),  "1"(b), "r"(m.d), "r"(m.v), "c"(m.s)
        : "%r8", "%r9", "%r10");
        /* q = quotient */
        return r;
}

#define UINT64 unsigned long long

typedef struct {
	UINT64 s;	/* shift */
	UINT64 v;	/* reciprocal of d */
	UINT64 d0;	/* divisor shifted up */
	UINT64 d1;
} recint;

recint recip1(UINT64  p);
UINT64 mulrec64(UINT64  a, UINT64  b, recint  v);


/******************************************************************************************/
/*  Zp utilities                                                                          */
/******************************************************************************************/

inline int add32s(int a, int b, int p) { int t; t = (a-p)+b; t += (t>>31) & p; return t; }
inline int sub32s(int a, int b, int p) { int t; t = (a-b);   t += (t>>31) & p; return t; }
inline int max32s(int a, int b) { if(a>b) return a; else return b; }
inline int min32s(int a, int b) { if(a<b) return a; else return b; }
inline LONG add64s(LONG a, LONG b, LONG p) { LONG t; t = (a-p)+b; t += (t>>63) & p; return t; }
inline LONG sub64s(LONG a, LONG b, LONG p) { LONG t; t = a-b; t += (t>>63) & p; return t; }
inline LONG mul64s(LONG a, LONG b, LONG p) {
        LONG q, r;
        __asm__ __volatile__(           \
        "       mulq    %%rdx           \n\t" \
        "       divq    %4              \n\t" \
        : "=a"(q), "=d"(r) : "0"(a), "1"(b), "rm"(p));
        return r;
}
inline LONG neg64s(LONG a, LONG p) { return (a==0) ? 0 : p-a; }
inline LONG submul64s(LONG a, LONG b, LONG c, LONG p) { return mul64s(sub64s(a,b,p),c,p); } // This is faster.
inline LONG muladd64s(LONG a, LONG b, LONG c, LONG p, recint P) { return add64s(mulrec64(a,b,P),c,p); }
inline LONG max64s(LONG a, LONG b) { if(a>b) return a; else return b; }
inline LONG min64s(LONG a, LONG b) { if(a<b) return a; else return b; }

/* c^(-1) mod p assuming 0 < c < p < 2^63 */
LONG modinv64s( LONG c, LONG p )
{   LONG d,r,q,r1,c1,d1;
    d = p;
    c1 = 1;
    d1 = 0;
    while( d != 0 ) {
        q = c / d;
        r = c - q*d; r1 = c1 - q*d1;
        c = d; c1 = d1;
        d = r; d1 = r1;
    }
    if( c!=1 ) return( 0 );
    if( c1 < 0 ) c1 += p;
    return( c1 );
}

/* a^n mod p assuming 0 <= a < p < 2^63 */
LONG powmod64s( LONG a, LONG n, LONG p )
{   LONG r,s;
    a += (a>>63) & p; // protect from bad input
    if( n==0 ) return 1;
    if( n==1 ) return a;
    for( r=1, s=a; n>0; n /= 2 ) { if( n & 1 ) r = mul64s(r,s,p); s = mul64s(s,s,p); }
    return r;
}

/* a^n mod p assuming 0 <= a < p < 2^63 */
LONG powmodrec64s( LONG a, LONG n, LONG p, recint P )
{   LONG r,s;
    a += (a>>63) & p; // protect from bad input
    if( n==0 ) return 1;
    if( n==1 ) return a;
    for( r=1, s=a; n>0; n /= 2 ) { if( n & 1 ) r = mulrec64(r,s,P); s = mulrec64(s,s,P); }
    return r;
}


//------------------------------------------------------------------------------
// The following aren't used by dlog64sceil.c (or anywhere else).
// TODO: Replace these with setupChrem64s(), chrem64s(), and logNaive64s().
//------------------------------------------------------------------------------
///* Get inverses in I[1], I[2], ..., I[n-1] for crt */
//void getinv32s( int n, int *M, int *I )
//{   int k,m,p;
//    m = M[0];
//    for( k=1; k<n; k++ ) { p = M[k]; I[k] = modinv64s(m % p, p); m *= p; }
//    return;
//}
//
//int modcrt32s( int n, int *U, int *M, int* I )
//{   int k,v,m,u,p,t,delta;
//    v = U[0]; m = M[0];
//    for( k=1; k<n; k++ ) {
//        u = U[k]; p = M[k];
//        delta = I[k]*(u-v) % p;
//        delta += (delta>>31) & p;
//        v = v+delta*m;
//        m = p*m;
//    }
//    return v;
//}
//
///* x = log[alpha](a) in Zp where alpha has order n */
//int disclog32s( int a, int alpha, int n, int p )
//{
//    int x; LONG y;
//    a = a % p; a += (a>>31) & p; // protect from bad input
//    for( y=1, x=0; x<n; x++ ) {
//       if( y==a ) return x;
//       y = (y*alpha) % p;
//    }
//    return -1;
//}
//
///* Compute sqrt(n) for 0<=n<2^31 without overflowing when n is close to 2^31 */
//int floorsqrt32s( int n )
//{
//   int s;
//   if( n<1 ) return 0;
//   for( s=0; n-s*s-2*s>0; s++ );
//   return s;
//}

int ceilsqrt32s( int n )
{
   int s;
   if( n<1 ) return 0;
   for( s=0; n-s*s-2*s-1>0; s++ );
   return s+1;
}

//------------------------------------------------------------------------------
// The following aren't used by dlog64sceil.c (or anywhere else).
// TODO: Replace these with hashShanks64s() and logShanks64s().
//------------------------------------------------------------------------------
//int ShanksHash32s( int alpha, int n, int *Y, int *X, int p )
//{
//   int m,M,i,x,y,z,h;
//   m = ceilsqrt32s(n);
//   for( y=alpha,i=1; i<m; i++ ) y = mul64s(alpha,y,p);
//   M = 2*m+1; // size of hash tables Y and X; M = 2m+1 means they'll be half full
//   for( i=0; i<M; i++ ) Y[i] = 0;
//   Y[1] = 1;
//   X[1] = 0;
//   for( z=y, x=m; x<n; x+=m ) {
//      for( h=z%M; Y[h]; h=add32s(h,1,M) );
//      Y[h] = z;
//      X[h] = x;
//      z = mul64s(y,z,p);
//   }
//   return M;
//}
//
//int logshanks( int y, int alpha, int *Y, int *X, int M, int n, int p )
//{  // output x = log[alpha](y) in Zp where alpha has order n
//   int i,h,x,z;
//   for( i=0; i<M/2; i++ ) {
//        for( h=y%M; Y[h]; h=add32s(h,1,M) )
//             if( Y[h]==y ) { x = sub32s(X[h],i,n); return x; }
//        y = mul64s(alpha,y,p);
//   }
//   return -1; // it's not there
//}

int MillerRabinTest64s(LONG n, int a )
{
        LONG m,b;
        int i,k;
        m = n-1;
        for( k=0; !(m&1); k++ ) m = m/2;

        b = powmod64s(a,m,n);

        if(b==1) return 1;
        for( i=0; i<k; i++ ) {
             if( b==n-1 ) return 1;
             b = mul64s(b,b,n);
        }
        return 0;
}

int isprime64s(LONG n)
{
        if( n<2 ) return 0;
        if( n==2 || n==3 ) return 1;
        if( n%2==0 || n%3==0 ) return 0;
        if( !MillerRabinTest64s( n, 2 ) ) return 0;
        if( !MillerRabinTest64s( n, 325 ) ) return 0;
        if( !MillerRabinTest64s( n, 9375 ) ) return 0;
        if( !MillerRabinTest64s( n, 28178 ) ) return 0;
        if( !MillerRabinTest64s( n, 450775 ) ) return 0;
        if( !MillerRabinTest64s( n, 9780504 ) ) return 0;
        if( !MillerRabinTest64s( n, 1795265022) ) return 0;
        return 1;
}

LONG prevFourierprime64s( LONG x, int k ) {
// compute the first prime p<x s.t. 2^k|p-1  if it exists
    LONG g,m,p;
    if( k>62 ) return 0; else g = 1LL << k; // g = 2^k
    for( m=(x-2)/g; m>0; m-- ) { p = m*g+1; if( isprime64s(p) ) return p; }
    return 0;
}

/******************************************************************************************/
/* Polynomial utilities                                                                   */
/******************************************************************************************/

void vecfill64s( LONG x, LONG *A, int n )
{   int i;
    for( i=0; i<n; i++ ) A[i] = x;
    return;
}

void polcopy64s( LONG *A, int d, LONG *B )
{   int i;
    for( i=0; i<=d; i++) B[i]=A[i];
    return;
}

/* print an array in form [a0,a1,...,an-1] */
void vecprint64s( LONG *A, int n )
{   int i;
    printf("[");
    for( i=0; i<n; i++ ) { printf("%lld",A[i]); if( i<n-1 ) printf(", "); }
    printf("];\n");
    return;
}

/* print an array [a0,a1,...,ad] in form ad*x^d+...+a1*x+a0 */
void polprint64s( LONG *A, int d ) {
    int i;
    if( d==-1 ) { printf("0;\n"); return; }
    for( i=d; i>0; i-- ) if( A[i]!=0 ) printf("%lld*x^%d+",A[i],i);
    printf("%lld;\n",A[0]);
    return;
}

/* for C = [c1,c2,...,cn] and X = [e1,e2,...,en] print  c1*x^e1+...+cn*x^en;  */
void polsparseprint64s( LONG *C, LONG *X, int n ) {
    int i;
    if( n==0 ) { printf("0;\n"); return; }
    for( i=n-1; i>0; i-- ) printf("%lld*x^%lld+",C[i],X[i]);
    printf("%lld*x^%lld;\n",C[i],X[i]);
    return;
}

void pol3sparseprint64s( LONG *C, LONG *X, int t ) {
    int i,dx,dy,dz,mask;
    LONG m;
    if( t==0 ) { printf("0;\n"); return; }
    mask = (1<<21)-1;
    for( i=0; i<t; i++ ) {
        m = X[i]; dz = m & mask; m = m>>21; dy = m & mask; dx = m>>21;
        printf("%lld*x^%d*y^%d*z^%d",C[i],dx,dy,dz);
        if( i!=t-1 ) printf("+"); else printf(";\n");
    }
    return;
}

void polnsparseprint64s( LONG *C, LONG *X, int t, int n ) {
    int numbits,i,j;
    LONG m,d,mask,D[63];
    if( t==0 ) { printf("0;\n"); return; }
    numbits = 63/n;
    mask = (1LL<<numbits)-1;
    for( i=0; i<t; i++ ) {
        m = X[i];
        printf("%lld",C[i]);
        for( j=0; j<n; j++ ) {
            D[j] = m & mask;
            m = m >> numbits;
        }
        for( j=1; j<=n; j++ ) {
            d = D[n-j];
            if( d!=0 ) printf("*x[%d]^%lld",j,d);
        }
        if( i!=t-1 ) printf("+"); else printf(";\n");
    }
    return;
}

void pol2sort64s( LONG *C, LONG *X, int n ) {
    LONG x,c;
    if( n<2 ) return;
    n--;
    x = X[n];
    c = C[n];
    pol2sort64s( C, X, n );
    while( n>0 && X[n-1]>x ) { X[n] = X[n-1]; C[n] = C[n-1]; n--; }
    X[n] = x;
    C[n] = c;
    return;
}

int vec2partition64s(LONG *D, LONG *C, int n ) {
    int i,j,m;
    LONG x,t,c;
    m = n/2;
    x = D[m]; D[m] = D[0];
    c = C[m]; C[m] = C[0];
    for( i=1,j=n-1; j>=i ; ) {
       if( D[i]<x ) { D[i-1] = D[i]; C[i-1] = C[i]; i++; }
       else { t = D[i]; D[i] = D[j]; D[j] = t; t = C[i]; C[i] = C[j]; C[j] = t; j--; }
    }
    D[i-1] = x;
    C[i-1] = c;
    return i-1;
}

void sort2vec64s(LONG *D, LONG *C, int n ) {
    int i;
    LONG x,c;
    if( n<2 ) return;
    if( n>50 ) { i = vec2partition64s(D, C, n);
                  sort2vec64s(D,C,i);
                  sort2vec64s(D+i+1,C+i+1,n-i-1);
                  //printf("QUICKSORT\n");
                  return;
                 }
    sort2vec64s(D,C,n-1);
    x = D[n-1];
    c = C[n-1];
    for( i=n-1; i>0 && x<D[i-1]; i-- ) { D[i] = D[i-1]; C[i] = C[i-1]; }
    D[i] = x;
    C[i] = c;
    return;
}

LONG pol2simp64s( LONG *C, LONG *X, int n, LONG p ) {
    LONG i,j,k,c,x;
    //pol2sort64s(C,X,n);
    sort2vec64s(X,C,n);
    k = 0;
    i = 0;
    for( k=0,i=0; i<n; k++ ) {
       c = C[i];
       x = X[i];
       for( j=i+1; j<n && X[j]==x; j++ ) c = add64s(c,C[j],p);
       C[k] = c; X[k] = x;
       i = j;
    }
    return k;
}

LONG pol2simpNoSort64s( LONG *C, LONG *X, int n, LONG p ) {
    LONG i,j,k,c,x;
    k = 0;
    i = 0;
    for( k=0,i=0; i<n; k++ ) {
       c = C[i];
       x = X[i];
       for( j=i+1; j<n && X[j]==x; j++ ) c = add64s(c,C[j],p);
       C[k] = c; X[k] = x;
       i = j;
    }
    return k;
}

void polscamul64s( LONG x, LONG *A, int d, LONG p )
{
    int i;
    if( x==1 ) return;
    if( x==-1 ) for( i=0; i<=d; i++ ) A[i] = neg64s(A[i],p);
    else for( i=0; i<=d; i++ ) A[i] = mul64s(x,A[i],p);
    return;
}

void polscamulrec64s( LONG x, LONG *A, int d, LONG p, recint P )
{
    int i;
    if( x==1 ) return;
    if( x==-1 ) for( i=0; i<=d; i++ ) A[i] = neg64s(A[i],p);
    else for( i=0; i<=d; i++ ) A[i] = mulrec64(x,A[i],P);
    return;
}

/* make polynomial in A monic */
void monic64s( LONG *A, int d, LONG p ) {
    int i; LONG inv;
    recint P;
    P = recip1(p);
    if( d<0 || A[d]==1 ) return;
    inv = modinv64s(A[d],p);
    for( i=0; i<d; i++ ) A[i] = mulrec64(inv, A[i], P);
    A[d] = 1;
    return;
}

/* make powers in A = [1, alpha, alpha^2, ..., alpha^(n-1)] */
void vecpow64s( int n, LONG *A, LONG alpha, LONG p )
{
   int i;
   //M_UINT recip, shift;  // Not used yet.
   LONG a4;
   recint P;

   if( n<1 ) return;
   //init( (M_UINT) p, &recip, &shift );  // Not used yet.
   P = recip1(p);

   A[0] = 1;
   for( i=1; i<4 && i<n; i++ ) A[i] = mulrec64(alpha,A[i-1],P);
   if( n<4 ) return;

   a4 = mulrec64(alpha,A[3],P);
   for( i=4; i<n-3; i+=4 ) { A[i+0] = mulrec64(a4,A[i-4],P);
                             A[i+1] = mulrec64(a4,A[i-3],P);
                             A[i+2] = mulrec64(a4,A[i-2],P);
                             A[i+3] = mulrec64(a4,A[i-1],P); }
   for( ; i<n; i++ ) A[i] = mulrec64(alpha,A[i-1],P);
   return;
}

/******************************************************************************************/
/* Polynomial routines                                                                    */
/******************************************************************************************/

LONG poleval64s(LONG *a, int d, LONG x, LONG p) {
    int i;
    LONG r;
    recint P;
    if( d==-1 ) return 0;
    P = recip1(p);
    for( r=a[d],i=d-1; i>=0; i-- ) r = muladd64s(r, x, a[i], p, P);
    return r;
}

int polnsparsedeg64s( LONG *X, int t, int n, int xi ) {
    int numbits,j,d;
    LONG x,mask,shift;
    // degree(f,x[i])
    if( xi<1 || xi>n ) return -1;
    numbits = 63/n;
    mask = (1LL<<numbits)-1;
    shift = numbits*(n-xi);
    d = -1;
    for( j=0; j<t; j++ ) {
        x = X[j] >> shift;
        if( xi>1 ) x = x & mask;
        if( x>d ) d = x;
    }
    return d;
}


void polnsparsevecdeg64s( LONG *X, int t, int n, int *D ) {
    int numbits,i,j,d,e;
    LONG x,y,mask,shift;
    // D[i] = degree(f,x[i]) for 0<=i<n with variables read from the right.
    for(i=0; i<n; i++ ) D[i] = 0;
    numbits = 63/n;
    mask = (1LL<<numbits)-1;
    for( j=0; j<t; j++ ) {
        x = X[j];
        for( i=0; i<n; i++ ) {
            if( i<n-1 ) d = x & mask; else d = x;
            if( d>D[i] ) D[i] = d;
            x = x >> numbits;
        }
    }
    return;
}



LONG polnsparseeval64s(LONG *C, LONG *X, int t, LONG *A, int n, LONG p ) {
// Evaluate f(A[0],A[1],...,A[n-1]) mod p where f(x1,x2,...,xn) = sum( c[i]*X[i], i=0..t-1 )
    int numbits,i,j;
    LONG mask,s,x,u;
    recint P;
    P = recip1(p);
    numbits = 63/n;
    mask = (1LL<<numbits)-1;
    s = 0;
    for( i=0; i<t; i++ ) {
        x = X[i];
        u = C[i];
        for( j=0; j<n; j++ ) { u = mulrec64(u,powmodrec64s(A[j],x&mask,p,P),P); x = x >> numbits; }
        s = add64s(s,u,p);
    }
    return s;
}

// Evaluate g(x1) = f(x1,A[0],A[1],...,A[n-2]) where f = sum( A[i]*X[i], i=0..t-1 )
// where g(x1) = sum( B[i]*Y[i], i=0..d ) and return d;
int poln2nm1sparseeval64s( LONG *C, LONG *X, int t, LONG *A, int n, LONG *B, LONG *Y, LONG p )
{
    int numbits,i,j,k;
    LONG mask,s,x,u;
    recint P;
    P = recip1(p);
    numbits = 63/n;
    mask = (1LL<<numbits)-1;
    s = numbits*(n-1);
    mask = (1LL<<s) - 1;

    for( i=0; i<t; i++ ) {
        x = X[i];
        u = C[i];
        for( j=0; j<n-1; j++ ) { u = mulrec64(u,powmodrec64s(A[j],x&mask,p,P),P); x = x >> numbits; }
        Y[i] = x;
        B[i] = u;
    }
    sort2vec64s(Y,B,t);
    k=0;
    i=0;
    while( i<t ) {
       s = B[i];
       for( j=i+1; j<t && Y[i]==Y[j]; j++ ) s = add64s(B[j],s,p);
       if( s!= 0 ) { B[k] = s; Y[k] = Y[i]; k++; }
       i = j;
    }
    return k;
}

// So beta = alpha^i and we evaluate f(x1,y) at y=beta
int poln2nm1kroneval64s( LONG *C, LONG *X, int t, LONG beta, int n, LONG *B, LONG *Y, LONG p )
{
    int numbits,i,j,k;
    LONG pow[63],mask,s,x,u;
    recint P;
    P = recip1(p);
    numbits = 63/n;
    s = numbits*(n-1);
    mask = (1LL<<s) - 1;

    pow[0] = 1;
    pow[1] = beta;
    for( i=2; i<s; i++ ) pow[i] = mulrec64(pow[i-1],pow[i-1],P);

    for( i=0; i<t; i++ ) {
        x = X[i];
        Y[i] = x >> s;
        x = x & mask;
        //u = powmod64s(beta,x,p);
        for( u=1,j=1; x>0; x=x>>1,j++ ) if( x&1 ) u = mulrec64(u,pow[j],P);
        B[i] = mulrec64(C[i],u,P);
    }

    sort2vec64s(Y,B,t);
    k=0;
    i=0;
    while( i<t ) {
       s = B[i];
       for( j=i+1; j<t && Y[i]==Y[j]; j++ ) s = add64s(B[j],s,p);
       if( s!= 0 ) { B[k] = s; Y[k] = Y[i]; k++; }
       i = j;
    }
    return k;
}

void randpoly64s( int n, int *DEG, int t, LONG *A, LONG *X, LONG p )
{   // monic in x[n-1]
    int numbits,i,j;
    LONG x,c,y;
    numbits = 63/n;
    for( i=0; i<t; i++ ) {
        for( c=0; c==0; c = rand64s(p) );
        A[i] = c;
        x = rand64s(DEG[0]);
        for( j=1; j<n; j++ ) { y = rand64s(DEG[j]+1); x = (x << numbits) | y; }
        X[i] = x;
    }
    // make it monic in x[0] and not be divisible by x[0]
    A[t-1] = 1;
    X[t-1] = (LONG)DEG[0] << (numbits*(n-1));
    X[0  ] = X[0] & ( (1LL << numbits*(n-1)) - 1 );
}

void randpolyNOTmonic64s( int n, int *DEG, int t, LONG *A, LONG *X, LONG p )
{
    int numbits,i,j;
    LONG x,c,y;
    numbits = 63/n;
    for( i=0; i<t; i++ ) {
        for( c=0; c==0; c = rand64s(p) );
        A[i] = c;
        x = rand64s(DEG[0]+1);
        for( j=1; j<n; j++ ) { y = rand64s(DEG[j]+1); x = (x << numbits) | y; }
        X[i] = x;
    }
    // make it have a constant
    X[0] = 0;
}

void randpolyNOTmonic64s2( int n, int *DEG, int D, int t, LONG *A, LONG *X, LONG p )
{   // with total degree at most D
    int numbits,i,j;
    LONG x,c,y,d;
    numbits = 63/n;
    for( i=0; i<t; i++ ) {
        for( c=0; c==0; c = rand64s(p) );
        A[i] = c;
        x = rand64s(DEG[0]+1);
        d = x;
        for( d=x,j=1; j<n; j++,d+=y ) { y = rand64s(DEG[j]+1); x = (x << numbits) | y; }
        if( d<=D ) X[i] = x; else i--;
    }
    // make it have a constant
    X[0  ] = 0;
    return;
}

LONG polsparseeval64s(LONG *c, LONG *a, int n, LONG x, LONG p) {
    int i;
    LONG d, y,r;
    recint P;
    if( n==0 ) return 0;
    P = recip1(p);
    r = c[n-1];
    for( i=n-1; i>0; i-- ) {
        d = a[i]-a[i-1];
        if( d<=0 ) return 0; // should not happen
        if( d==1 ) y = x; else y = powmodrec64s(x,d,p,P);
        //r = (y * r + c[i-1]) % p;
        r = muladd64s(y,r,c[i-1],p,P);
    }
    y = powmodrec64s(x,a[0],p,P);
    r = mulrec64(y,r,P);
    return r;
}

int poladd64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG p) {
// c = a + b mod p
   int i,m;
   m = min32s(da,db);
   for( i=0; i<=m; i++ ) c[i] = add64s(a[i],b[i],p);
   if( da==db ) { while( da>=0 && c[da]==0 ) da--; return da; }
   if( da<db ) { for( i=da+1; i<=db; i++ ) c[i] = b[i]; return db; }
   for( i=db+1; i<=da; i++ ) c[i] = a[i]; return da;
}

int polsub64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG p) {
// c = a-b mod p
   int i,m;
   m = min32s(da,db);
   //if( da<0 ) printf("SUB da=%d\n",da);
   //if( db<0 ) printf("SUB db=%d\n",db);
   for( i=0; i<=m; i++ ) c[i] = sub64s(a[i],b[i],p);
   if( da==db ) { while( da>=0 && c[da]==0 ) da--; return da; }
   if( da>db ) { for ( i=db+1; i<=da; i++ ) c[i] = a[i]; return da; }
   else { for( i=da+1; i<=db; i++ ) c[i] = neg64s(b[i],p); return db; }
}

/* compute C(x) = A(x)^2 mod p and return deg(C) */
/* we allow C to overwrite A i.e. polsqr64s(A,A,d,p) */
int polsqr64s( LONG * A, LONG * C, int d, LONG p )
{
    int i,k,m; ULONG z[2];
    if( d<0 ) return d;
    for( k=2*d; k>=0; k-- ) {
       m = min32s(k,d);
       i = max32s(0,k-d);
       z[0] ^= z[0]; // = 0?
       z[1] ^= z[1]; // = 0?
       while( i<m-2 ) {
            zfma(z,A[i++],A[m--]);
            if( z[1]>=p ) z[1] -= p;
            zfma(z,A[i++],A[m--]);
       }
       if( i<m ) {
            zfma(z,A[i++],A[m--]);
            if( z[1]>=p ) z[1] -= p;
       }
       zadd(z,z[0],z[1]);
       if( z[1]>=p ) z[1] -= p;
       if( i==m ) zfma(z,A[i],A[i]);
       zmod(z,p);
       C[k] = z[0];
    }
    //for( dc = 2*d; dc>=0 && C[dc]==0; dc-- );
    // Why is this loop here? Z_p has no zero-divisors.
    return( 2*d );
}

/* compute C(x) = A(x) * B(x) mod p and return deg(C) */
/* we allow C to overwrite either A or B i.e. polmul64s(A,B,A,da,db,p) */
int polmul64s( LONG * A, LONG * B, LONG * C, int da, int db, LONG p)
{
    int i,k,m; ULONG z[2];
    if( da<0 || db<0 ) return da;
    int dc = da+db;
    for( k=dc; k>=0; k-- ) {
       i = max32s(0,k-db);
       m = min32s(k,da);
       z[0] ^= z[0]; // = 0?
       z[1] ^= z[1]; // = 0?
       while( i<m ) {
           zfma(z,A[i],B[k-i]); i++;
           if( z[1]>=p ) z[1] -= p;
           zfma(z,A[i],B[k-i]); i++;
       }
       if( i==m ) {
           zfma(z,A[i],B[k-i]);
           if( z[1]>=p ) z[1] -= p;
       }
       zmod(z,p);
       C[k] = z[0];
    }
    return( dc );
}

/* divide A by B and put the remainder and quotient in A */
/* return the degree of the remainder                    */
int poldiv64s( LONG * A, LONG * B, int da, int db, LONG p )
{
    int dq,dr,k,j,m; LONG t,inv; ULONG z[2];
    if( db<0 ) { printf("division by zero\n"); exit(1); }
    if( da<db ) return da; else { dq = da-db; dr = db-1; }
    inv = modinv64s(B[db],p);
    for( k=da; k>=0; k-- ) {
        z[0] = 0ll; z[1] = 0ll;
        m = min32s(dr,k);
        j = max32s(0,k-dq);
        //for( j=max32s(0,k-dq); j<=m; j++ ) { t -= ((LONG) B[j])*A[k-j+db]; t += (t>>63) & M; }
        while( j<m ) {
            zfma(z,B[j],A[k-j+db]); j++;
            if( z[1]>=p ) z[1] -= p;
            zfma(z,B[j],A[k-j+db]); j++;
        }
        if( j==m ) zfma(z,B[j],A[k-j+db]);
        if( z[1]>=p ) z[1] -= p;
        zmod(z,p);
        t = A[k] - z[0];
        t += (t>>63) & p;
        if( k>=db && inv!=1 ) t = mul64s(t,inv,p);
        A[k] = t;
    }
    while( dr>=0 && A[dr]==0 ) dr--;
    return( dr );
}

void pollongdiv64s( LONG *f, LONG *g, int df, int dg, LONG *y, int n, LONG p )
{
    int i, j, k;
    __int128_t t,M;

    if( g[0]==0 ) { printf("no inverse\n"); return; }

    M = (__int128_t) p << 64;

    y[0] = f[0];//f[0]*modinv64s(g[0],p);
    for( i=1; i<=df && i<n; i++ )
    {
        t = (M-p)+f[i];
        for( j=0; j<dg, j<i; j++ )
        {
            t = t - (__int128_t) y[j]*g[i-j];
            t += (t>>127) & M;
        }
        y[i] = t % p;
    }
    for( ; i<=dg && i<n; i++ )
    {
        t = M;
        for( j=0; j<dg, j<i; j++ )
        {
            t = t - (__int128_t) y[j]*g[i-j];
            t += (t>>127) & M;
        }
        y[i] = t % p;
    }
    for( ; i<n; i++ )
    {
        t = M;
        for( j=1; j<=dg; j++ )
        {
            t = t - (__int128_t) y[i-j]*g[j];
            t += (t>>127) & M;
        }
        y[i] = t % p;
    }
}

void polinvmod64s( LONG *f, int n, LONG p, LONG *y, LONG *T )
{
    // compute y = f^(-1) mod p as a power series to O(x^n)
    // T needs to be of size ceil(3n/2)
    int m,i;
    if( f[0]==0 ) { printf("no inverse\n"); return; }
    if( n==1 ) { y[0] = modinv64s(f[0],p); return; }
    m = (n+1)/2;
    polinvmod64s( f, m, p, y, T );
    polmul64s( y, f, T, m-1, n-1, p );
    for( i=0; i<m; i++ ) T[i] = neg64s(T[i+m],p);
    polmul64s( y, T, T, m-1, m-1, p );
    for( i=0; i<m && i+m<n; i++ ) y[i+m] = T[i];
    return;
}

/* compute gcd(A,B) and put gcd in A and return it's degree */
int polsubmulP( LONG *A, LONG *B, LONG a, LONG b, int dA, int dB, LONG p, recint P ) {

   // compute A = A - (ax+b) B efficiently

   LONG s,t; int i, d;

   if( dB==-1 ) return dA; // B = 0

   d = dA;

   // if deg(A) <= deg(B) then pad A with zeroes
   while( dA<=dB ) A[++dA] = 0;

   // constant term is special
   t = mulrec64(b,B[0],P);
   A[0] = sub64s(A[0],t,p);

   //for( i=1; i<=dB; i++ ) { t = mul64s(a,B[i-1],p); t = add64s(t,mul64s(b,B[i],p),p); A[i] = sub64s(A[i],t,p); }
   for( i=1; i<=dB; i++ ) { t = mulrec64(a,B[i-1],P); t = add64s(t,mulrec64(b,B[i],P),p); A[i] = sub64s(A[i],t,p); }

   // update leading term from B
   t = mulrec64(a,B[dB],P);
   A[dB+1] = sub64s(A[dB+1],t,p);

   // compute and return degree
   while( dA>=0 && (A[dA]==0 || A[dA]==p) ) dA--;

   if( dA==d ) printf("polsubmul failure: dAin=%d dAout=%d\n",d,dA);
   return dA;
}

int polsubmul( LONG *A, LONG *B, LONG a, LONG b, int dA, int dB,
               LONG p, M_UINT recip, M_UINT shift ) {
   // compute A = A - (ax+b) B efficiently

   LONG t; int i; ULONG z[2];

   if( dB==-1 ) return dA; // B = 0
   z[0] = z[1] = 0LL;

   // if deg(A) <= deg(B) then pad A with zeroes
   while( dA<=dB ) A[++dA] = 0;

   // constant term is special
   t = mul64s(b,B[0],p) ;
   A[0] = sub64s(A[0],t,p);

   for( i=1; i<=dB; i++ ) { zmul(z,a,B[i-1]); zfma(z,b,B[i]); zmod(z,p); t = A[i]-z[0]; A[i] = t + ((t>>63)&p); }

   // update leading term from B
   t = mul64s(a,B[dB],p);
   A[dB+1] = sub64s(A[dB+1],t,p);

   // compute and return degree
   while( dA>=0 && (A[dA]==0 || A[dA]==p) ) dA--;
   return dA;
}


/* compute gcd(A,B) and put gcd in A and return it's degree */
/* Both A and B are destroyed */
int polgcd64s( LONG * A, LONG * B, int da, int db, LONG p ) {
    int dr; LONG *C, *D, *R, u, a, b;
    recint P;
    if( db<0 ) { printf("division by zero\n"); exit(1); }
    P = recip1(p);
    C = A; D = B;
    if( da<db ) { R = C; C = D; D = R; dr = da; da = db; db = dr; }
    while( 1 ) {
        if( db>0 && da-db==1 ) { // normal case
        //if( 0 ) {
            u = modinv64s(D[db],p);
            a = mulrec64(C[da],u,P);
            b = mulrec64(a,D[db-1],P);
            b = mulrec64(u,sub64s(C[da-1],b,p),P);  // quotient = a x + b
            dr = polsubmulP(C,D,a,b,da,db,p,P);  // C = C - (a x + b) D
            //dr = polsubmul(C,D,a,b,da,db,p,1,1);  // C = C - (a x + b) D
        }
        else dr = poldiv64s(C,D,da,db,p);
        if( dr<0 ) { /* D|C so gcd(A,B)=D */
            if( D!=A ) polcopy64s(D,db,A);
            monic64s( A, db, p );
            return db;
        }
        R = C; C = D; D = R; da = db; db = dr;
        //printf("da=%d db=%d\n",da,db);
    }
}

void polgcdext64s( LONG *A, LONG *B, int da, int db,
                  LONG *G, LONG *S, LONG *T, int *dG, int *dS, int *dT,
                  LONG *W, LONG p )
{
    // Solve S A + T B = G = monic gcd(A,B) for G,S,T in Zp[x]
    // The arrays A and B are used for the remainder sequence so are destroyed
    // G,S,T must be of size max(da+1,db+1)
    // W is working storage of size 2max(da+1,db+1)
    // if S==0 or T==0 then S (and/or T) are not computed

    int m,dr,ds,dt,dq,ds1,ds2,dt1,dt2; LONG a,b,u;
    LONG *q,*r,*r1,*r2,*s,*s1,*s2,*t,*t1,*t2;

    //P_INT P = modinit(p);
    recint P = recip1(p);

    if( da<0 || db<0 ) { printf("inputs must be non-zero\n"); exit(1); }
    m = max32s(da+1,db+1);
    r1 = A; r2 = B;
    if(S) { s1 = S; s2 = W;   s1[0]=1; ds1=0; ds2=-1; }
    if(T) { t1 = T; t2 = W+m; t2[0]=1; dt2=0; dt1=-1; }
    //M_UINT recip, shift;
    //init( (M_UINT) p, &recip, &shift );
    M_UINT recip = 1; M_UINT shift = 1;
    while( 1 ) {
        if( db>0 && da-db==1 ) { // normal case
            u = modinv64s(r2[db],p);
            a = mulrec64(r1[da],u,P);
            b = mulrec64(a,r2[db-1],P);
            b = mulrec64(u,sub64s(r1[da-1],b,p),P);                       // quotient = a x + b
            dr = polsubmul(r1,r2,a,b,da,db,p,recip,shift);              // r1 = r1 - (a x + b) r2
            if(S) ds = polsubmul(s1,s2,a,b,ds1,ds2,p,recip,shift);      // s1 = s1 - (a x + b) s2
            if(T) dt = polsubmul(t1,t2,a,b,dt1,dt2,p,recip,shift);      // t1 = t1 - (a x + b) t2
        }
        else {
            dr = poldiv64s(r1,r2,da,db,p);                 // r1 = [remainder,quotient]
            q  = r1+db; dq = da-db;
            if(S) ds = polmul64s(q,s2,G,dq,ds2,p);
            if(S) ds = polsub64s(s1,G,s1,ds1,ds,p);        // s1 = s1 - q s2
            if(T) dt = polmul64s(q,t2,G,dq,dt2,p);
            if(T) dt = polsub64s(t1,G,t1,dt1,dt,p);        // t1 = t1 - q t2
        }
        if( dr<0 ) { /* D|C so gcd(A,B)=D */
            polcopy64s(r2,db,G);
            if(S) if( s2!=S ) polcopy64s(s2,ds2,S);
            if(T) if( t2!=T ) polcopy64s(t2,dt2,T);
            if( G[db]!=1 ) {
                u = modinv64s(G[db],p);
                polscamulrec64s(u,G,db,p,P);
                if(S) polscamulrec64s(u,S,ds2,p,P);
                if(T) polscamulrec64s(u,T,dt2,p,P);
            }
            dG[0] = db;
            if(S) dS[0] = ds2;
            if(T) dT[0] = dt2;
            return;
        }
        r = r1; r1 = r2; r2 = r;  da = db;   db = dr;
        if(S) { s = s1; s1 = s2; s2 = s; ds1 = ds2; ds2 = ds; }
        if(T) { t = t1; t1 = t2; t2 = t; dt1 = dt2; dt2 = dt; }
    }
}

int FFTpolpowmod64s( LONG * A, LONG n, LONG * B, int da, int db, LONG *C, LONG *T, LONG p );
//int sinterp(int *A, int *X, int t, int n, int DegBnd, int alpha, int p );

/* C(x) := A(x)^n mod B(x) mod p;  0<=deg(A)<deg(B) and R must be of size 2*db-1 */
/* If A(x) is not reduced mod B(x) then we first compute C(x) := A(x) mod B(x)   */
int polpowmod64s( LONG * A, LONG n, LONG * B, int da, int db, LONG *C, LONG *R, LONG p )
{
    if( n==0 ) { C[0] = 1; return 0; }
    if( da>=db ) da = poldiv64s(A,B,da,db,p);                   // reduce A mod B first

if( db<600 ) {

    int dc,k,b[63];
    for( k=0; n>0; k++ ) { b[k]=n&1; n=n/2; }
    polcopy64s(A,da,C);
    dc = da;
    k--;
    while( k>0 ) { k--;
       // Main step: compute C := C^2 mod B in Zp[x]
       dc = polmul64s(C,C,R,dc,dc,p);                           //printf("deg(R) = %d; R = ",dc); polprint64s(R,dc);
       dc = poldiv64s(R,B,dc,db,p);
       polcopy64s(R,dc,C);                                      //printf("deg(C) = %d; C = ",dc); polprint64s(C,dc);
       if( b[k]==1 ) {                                          //printf(" b[%d]=%d \n", k, b[k] );
           dc = polmul64s(A,C,R,da,dc,p);                       //printf("deg(R) = %d; R = ",dc); polprint64s(R,dc);
           dc = poldiv64s(R,B,dc,db,p);
           polcopy64s(R,dc,C);                                  //printf("deg(C) = %d; C = ",dc); polprint64s(C,dc);
       }
    }
    return dc;

} else {
    int N,dc;
    LONG *T;
    for( N=1; N<2*db-1; N*=2 );
    /************ GLOBAL alpha */
    //printf("    Splitting deg(f)=%d using FFT with n=%d\n",db,N);
    T = array(6*N);
    dc = FFTpolpowmod64s(A,n,B,da,db,C,T,p);
    free(T);
    return dc;
    }
}

// Input f in Zp[x] of degree d > 0, a known product of d linear factors.
// Output roots of f in R.
// The input array f is destroyed.
// W is a scratch array of size at least 3*d
void polsplit64s( LONG *f, int d, LONG *R, LONG *W, LONG p )
{
   int da,dg; LONG alpha, A[2];
   if( d==1 ) { alpha = p-f[0]; R[0] = alpha; return; }
   alpha = rand64s(p); A[1] = 1; A[0] = alpha;
   da = polpowmod64s( A, (p-1)/2, f, 1, d, W, W+d, p );
   if( da==0 ) return polsplit64s(f,d,R,W,p);      // alpha is unlucky, try again
   W[0] = add64s(W[0],1,p);                        // W = (x+alpha)^((p-1)/2) + 1 mod f
   polcopy64s( f, d, W+d );
   dg = polgcd64s( W, W+d, da, d, p );             // g = gcd( W, f ) in W
   if( dg==0 ) return polsplit64s(f,d,R,W,p);      // g = 1 ==> alpha is unlucky, try again
   poldiv64s(f,W,d,dg,p);                          // compute quotient q = f/g destroying f
   polcopy64s(W,dg-1,f);                           // f = [ g mod x^dg followed by q ]
   polsplit64s(f+dg,d-dg,R,W,p);
   f[dg] = 1;
   polsplit64s(f,dg,R+d-dg,W,p);
   return;
}

/*******************************************************************************/

/* R := Roots(f) mod p */
int polroots64s( LONG * f, int d, LONG * R, LONG *W, LONG p )
{
   int i, da, dg; // printf("f := "); polprint64s(f,d);
   LONG A[2];
   for( i=0; i<d && f[i]==0; i++ );
   if( i>0 ) { R[0]=0; return( 1 + polroots64s(f+i,d-i,R+1,W,p) ); }
   if( f[d]!=1 ) monic64s(f,d,p);

   A[1] = 1;
   A[0] = 0;
   da = polpowmod64s( A, p-1, f, 1, d, W, (W+d), p );    // W = x^(p-1) mod f
   if( da==0 && W[0]==1 ) dg = d; // f is all linear factors
   else { W[0] = sub64s(W[0],1,p);
          dg = polgcd64s( f, W, d, da, p ); }          // f = gcd(f,W-1)
   //printf("g := "); polprint64s(f,dg);
   //printf("Roots: def(f)=%d  #roots=%d\n",d,dg);
   if( dg==0 ) return 0;
   polsplit64s( f, dg, R, W, p );
   return dg;
}

int polBerMas64s( LONG *a, int N, LONG *L, LONG *W, LONG p )
{
    // Input sequence a = [a1,a2,a3,...,aN]
    // Output polynomial Lambda(x) is written to L
    // Uses the half extended Euclidean algorithm
    int i,m,n,dr,dq,dr0,dr1,dv0,dv1,dt;
    LONG *r,*q,*r0,*r1,*v0,*v1,*t,u,A,b;
    while( N>0 && a[N-1]==0 ) N--; // ignore leading zeroes
    n = N/2;
    N = 2*n;
    if( N==0 ) return -1;
    m = N-1;
    // W is space for r0 = x^N and r1 of degree m and v0 and v1 of degree at most n
    r0 = W; r1 = r0+N+1; v0 = r1+N; v1 = v0+n+1;
    vecfill64s(0,r0,N); r0[N] = 1; dr0 = N;             // r0 = x^(2*n)
    for(i=0; i<N; i++) r1[i] = a[m-i];
    for(dr1=m; dr1>=0 && r1[dr1]==0; dr1--);            // r1 = sum(a[m-i]*x^i,i=0..m)
    if( dr1==-1 ) return -1;
    dv0 = -1;                                           // v0 = 0
    v1[0] = 1; dv1 = 0;                                 // v1 = 1
    recint P;
    P = recip1(p);
    while( n <= dr1 ) {
        if( dr1>0 && dr0-dr1==1 ) { // normal case
        //if( 0 ) {
            u = modinv64s(r1[dr1],p);
            A = mul64s(r0[dr0],u,p);
            b = mul64s(A,r1[dr1-1],p);
            b = mul64s(u,sub64s(r0[dr0-1],b,p),p);             // quotient q = A x + b
            dr = polsubmulP(r0,r1,A,b,dr0,dr1,p,P);            // r0 = r0 - (A x + b) r1
            //dr = polsubmul(r0,r1,A,b,dr0,dr1,p,1,1);            // r0 = r0 - (A x + b) r1
            dt = polsubmulP(v0,v1,A,b,dv0,dv1,p,P);            // v0 = v0 - (A x + b) v1
            //dt = polsubmul(v0,v1,A,b,dv0,dv1,p,1,1);            // v0 = v0 - (A x + b) v1
        } else {
           dr = poldiv64s(r0,r1,dr0,dr1,p);
           q = r0+dr1; dq = dr0-dr1;                           // q = quo(r0,r1)
           dt = polmul64s(q,v1,L,dq,dv1,p);
           dt = polsub64s(v0,L,v0,dv0,dt,p);
        }
        r = r0; r0 = r1; r1 = r; dr0 = dr1; dr1 = dr;         // r0,r1 = r1,rem(r0,r1)
        t = v0; v0 = v1; v1 = t; dv0 = dv1; dv1 = dt;         // v0,v1 = v1,v0 - q*v1
        //printf("r0 = "); polprint64s(r0,dr0);
        //printf("r1 = "); polprint64s(r1,dr1);
        //printf("v0 = "); polprint64s(v0,dv0);
        //printf("v1 = "); polprint64s(v1,dv1);
    }
    if( dv1>=0 ) {
        polcopy64s(v1,dv1,L);
        monic64s(L,dv1,p);
    }
    return dv1;
}


void polSolve64s( LONG *R, int n, LONG *L, LONG *a, LONG *X, LONG p )
{
    // R = [r1,r2,...,rn] is the roots of Lambda
    // L = Lambda(x) = product( x-R[i], i=1..n )
    // a = [a1,a2,...,an] is the right-hand-side
    // Solve V X = a for X where
    //
    //     [ 1        1        ... 1        ]
    //     [ r1       r2       ... rn       ]
    // V = [ r1^2     r2^2     ... rn^2     ]
    //     [ ...      ...      ... ...      ]
    //     [ r1^(n-1) r2^(n-1) ... rn^(n-1) ]
    //
    // Let Qj(x) = L(x)/(x-rj) = q1 + q2 x + ... + qn x^(n-1).
    // Let u = Qj(rj) and v = [q1,q2,...,qn=1] be the array of coefficients of Qj(x).
    // Then the solution Xj = v dot a / u for 1 <= j <= n.
    //
    // So one way to do this is to compute Qj(x) explicitly then evaluate Qj(rj).
    // That would need additional space for Qj(x) since we need to reuse L(x).
    // I suppose we could use L := L/(x-rj) then L := (x-rj)*L to recover L(x).
    // Intead I compute the coefficients qn,...,q2,q1 of Qj(x) in descending order
    // (in the variable q below) and assemble v dot a (in the variable z) and evaluate
    // Qj(rj) (in the variable u) simultaneously.

    int i,j;
    LONG q,r,u;
    ULONG z[2];
    recint P;

    P = recip1(p);
    for( j=0; j<n; j++ ) {
         r = R[j];                          // jth root of L
         z[0] = a[n-1];
         z[1] = 0;
         //z = a[n-1];
         u = 1;
         for( q=1,i=n-1; i>0; i-- ) {
              q = muladd64s(r,q,L[i],p,P); //q = (r*q+L[i]) % p;
              zfma(z,q,a[i-1]);
              if( z[1]>=p ) z[1] -= p;
              u = muladd64s(r,u,q,p,P);    //u = (r*u+q) % p;
         }
         zmod(z,p);                      // z = z % p
         u = modinv64s(u,p);
         X[j] = mulrec64(z[0],u,P);
         //X[j] = mul64s(z,u,p);
    }
    return;
}

int polsparsemul64s( LONG *A, LONG *X, int m, LONG *B, LONG *Y, int n, LONG p , LONG *C, LONG *Z) {
   int i,j,k,t;
   LONG s;
   recint P;
   P = recip1(p);
   for( i=0; i<m; i++ )
      for( j=0; j<n; j++ ) { Z[n*i+j] = X[i]+Y[j]; C[n*i+j] = mulrec64(A[i],B[j],P); }
   t = m*n;
   sort2vec64s(Z,C,t);
   k=0;
   i=0;
   while( i<t ) {
       s = C[i];
       for( j=i+1; j<t && Z[i]==Z[j]; j++ ) s = add64s(C[j],s,p);
       if( s!= 0 ) { C[k] = s; Z[k] = Z[i]; k++; }
       i = j;
   }
   return k;
}
