#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include "int128g.c"


#define LONG long long int

/********************************************************************************/
/*  Roman's assembler utilities                                                 */
/********************************************************************************/

#define WORDSIZE 64
typedef unsigned long ULONG;

/* [x0:x1] := a*b */
#define MUL211(x0,x1,a,b) __asm__(\
"       mulq    %3      \n\t"   \
: "=a"(x0), "=d"(x1) : "0"(a), "r"(b) : "cc")

/* x0 := [x0:x1] / a */
/* x1 := [x0:x1] % a */
#define DIV21H(x0,x1,a) __asm__(\
"       divq    %2      \n\t"   \
: "=a"(x0), "=d"(x1) : "r"(a), "0"(x0), "1"(x1) : "cc")



        /* z += a1:a0 */
        #define zadd(z,a0,a1) __asm__(\
        "       addq    %4, %0  \n\t" \
        "       adcq    %5, %1  \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]) : "0"(z[0]), "1"(z[1]), "r"(a0), "r"(a1))

        /* z -= a1:a0 */
        #define zsub(z,a0,a1) __asm__(\
        "       subq    %4, %0  \n\t" \
        "       sbbq    %5, %1  \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]) : "0"(z[0]), "1"(z[1]), "r"(a0), "r"(a1))

        /* z = a*b */
        #define zmul(z,a,b) __asm__(\
        "       mulq    %%rdx   \n\t" \
                : "=a"(z[0]), "=d"(z[1]) : "a"(a), "d"(b))

        /* z += a*b */
        #define zfma(z,a,b) do {        \
        unsigned long u,v;              \
        __asm__(                        \
        "       mulq    %%rdx           \n\t" \
        "       addq    %%rax, %0       \n\t" \
        "       adcq    %%rdx, %1       \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]), "=a"(u), "=d"(v) : "0"(z[0]), "1"(z[1]), "a"(a), "d"(b));\
        } while (0)

        /* z -= a*b */
        #define zfms(z,a,b) do {        \
        unsigned long u,v;              \
        __asm__(                        \
        "       mulq    %%rdx           \n\t" \
        "       subq    %%rax, %0       \n\t" \
        "       sbbq    %%rdx, %1       \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]), "=a"(u), "=d"(v) : "0"(z[0]), "1"(z[1]), "a"(a), "d"(b));\
        } while (0)

        /* z[0] = z % p */
        /* z[1] = z / p */
        /* quotient can overflow */
        #define zdiv(z,p) __asm__(\
        "       divq    %4      \n\t" \
                : "=a"(z[1]), "=d"(z[0]) : "a"(z[0]), "d"(z[1]), "r"(p))

        /* z = z % p safe */
        #define zmod(z,p) __asm__(\
        "       divq    %4      \n\t" \
        "       xorq    %0, %0  \n\t" \
                : "=a"(z[1]), "=d"(z[0]) : "a"(z[0]), "d"(z[1] < p ? z[1] : z[1] % p), "r"(p))

        /* z = z << s */
        #define zshl(z,s) __asm__(\
        "       shldq   %%cl, %0, %1    \n\t" \
        "       shlq    %%cl, %0        \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]) : "0"(z[0]), "1"(z[1]), "c"(s))


typedef struct P_INT P_INT;
struct P_INT {
        LONG p;        /* prime */
        LONG s;        /* shift */
        LONG d;        /* normalized */
        LONG v;        /* reciprocal */
};


/* initialize mod p struct */
inline P_INT modinit(LONG p)
{
        P_INT m;

        /* most significant bit location 63..0 */
        __asm__ __volatile__("bsrq %0, %0" : "=r"(m.s) : "0"(p));
        m.p = p;
        m.s = 64-m.s-1;
        m.d = p << m.s;

        /* reciprocal: (2^128-1)/n - 2^64 */
        /* [(2^64-n-1)*2^64 + (2^64-1)]/n */
        __asm__ __volatile__(\
        "       movq    %1, %%rdx       \n\t" \
        "       addq    $1, %%rdx       \n\t" \
        "       negq    %%rdx           \n\t" \
        "       movq    $-1, %%rax      \n\t" \
        "       divq    %1              \n\t" \
                : "=a"(m.v) : "r"(m.d) : "%rdx");
        return m;
}

inline LONG mulmodinv(LONG a, LONG b, P_INT m)
{
        LONG q, r;
        /* 0: compute  a*b = u1:u0, lshift by s, r9:r8 = (u1+1):u0 */
        /* 1: compute u1*v = q1:q0, q1:q0 += (u1+1):u0, in rdx:rax */
        /* 2: compute    r = u0-q1*d in r8 */
        /* 3: if r > q0 then q1 = q1-1, r = r+d (possible) */
        /* 4: if r >= d then q1 = q1+1, r = r-d (unlikely) */
        __asm__ __volatile__(\
        "0:     mulq    %%rdx                   \n\t" \
        "       shldq   %%cl,  %%rax, %%rdx     \n\t" \
        "       shlq    %%cl,  %%rax            \n\t" \
        "       leaq    1(%%rdx), %%r9          \n\t" \
        "       movq    %%rax, %%r8             \n\t" \
        "1:     movq    %%rdx, %%rax            \n\t" \
        "       mulq    %5                      \n\t" \
        "       addq    %%r8,  %%rax            \n\t" \
        "       adcq    %%r9,  %%rdx            \n\t" \
        "2:     imulq   %4,    %%rdx            \n\t" \
        "       subq    %%rdx, %%r8             \n\t" \
        "3:     cmpq    %%rax, %%r8             \n\t" \
        "       jbe     4f                      \n\t" \
        "       addq    %4,    %%r8             \n\t" \
        "       decq    %%rax                   \n\t" \
        "4:     cmpq    %4,    %%r8             \n\t" \
        "       jae     5f                      \n\t" \
        "       jmp     6f                      \n\t" \
        "5:     subq    %4,    %%r8             \n\t" \
        "       incq    %%rax                   \n\t" \
        "6:     shrq    %%cl,  %%r8             \n\t" \
        "       movq    %%r8,  %%rdx            \n\t" \
        : "=a"(q), "=d"(r)
        :  "0"(a),  "1"(b), "r"(m.d), "r"(m.v), "c"(m.s)
        : "%r8", "%r9", "%r10");
        /* q = quotient */
        return r;
}


/* floor of log2 */
ULONG ILOG2(ULONG x)
{
        ULONG y;
        for (y=0; x >>= 1; y++);
        return y;
}

/* reciprocal of full word */
ULONG RECIP(ULONG p)
{
        ULONG u0, u1;
        u1 = -(p+1);
        u0 = -1;
        DIV21H(u0,u1,p);
        return u0;
}

/* compute reciprocal and shift */
void INIT(ULONG d, ULONG *r, ULONG *s)
{
        int l;
        l = ILOG2(d);
        if( d != (ULONG)1 << l ) l++;
        d = d << (WORDSIZE-l);
        *r = RECIP(d);
        *s = l-1;
}

/* compute remainder via reciprocal */
inline ULONG REM(ULONG n, ULONG d, ULONG recip, ULONG s)
{
        ULONG u0,u1,q;
        MUL211(u0,u1,n,recip);
        q = (n - u1) >> 1;
        q = (q + u1) >> s;
        q = n - q*d;
        if( q!=d ) return q; else return 0;
}

/* compute remainder via reciprocal */
inline int MULMOD( ULONG a, ULONG b, ULONG p, ULONG recip, ULONG s)
{
        ULONG u0,u1,q,c;
        c = a * b;
        MUL211(u0,u1,c,recip);
        q = (c - u1) >> 1;
        q = (q + u1) >> s;
        c = c - q*p;
        if( (c!=p) ) return c; else return 0;
}

/********************************************************************************/
/*  Integer routines                                                            */
/********************************************************************************/

LONG POWMOD64s( LONG a, LONG n, recint P )
{
    /* a^n mod p assuming 0 <= a < p < 2^63 */
    ULONG r,s;
    if( n==0 ) return 1;
    if( n==1 ) return a;
    r = 1;
    s = a;
    while( 1 ) {
       //if( n&1 ) r = mul64s(r,s,p);
       if( n&1 ) r = mulrec64(r,s,P);
       n = n/2;
       //if( n>0 ) s = mul64s(s,s,p); else return r;
       if( n>0 ) s = mulrec64(s,s,P); else return r;
    }
}

/********************************************************************************/
/*  Vector routines                                                             */
/********************************************************************************/

void VECADD64s( LONG *A, LONG *B, LONG *C, int n, LONG p ) {
      int i;
      LONG t;
      for( i=0; i<n; i++ ) { t = A[i]-p+B[i]; t += (t>>63) & p; C[i] = t; }
      return;
}

void VECADDINP64s( LONG *A, LONG *B, int n, LONG p ) {
      int i;
      LONG t;
      for( i=0; i<n; i++ ) { t = A[i]-p+B[i]; t += (t>>63) & p; A[i] = t; }
      return;
}

void VECMUL64s( LONG *A, LONG *B, LONG *C, int n, LONG p ) {
      int i;
      recint P = recip1(p);
      for( i=0; i<n; i++ ) C[i] = mulrec64(A[i],B[i],P);
      return;
}

void VECMULINP64s( LONG *A, LONG *B, int n, LONG p ) {
      int i;
      recint P = recip1(p);
      for( i=0; i<n; i++ ) A[i] = mulrec64(A[i],B[i],P);
      return;
}

void VECSCAMUL64s( LONG x, LONG *A, int n, LONG p ) {
      int i;
      recint P = recip1(p);
      for( i=0; i<n; i++ ) A[i] = mulrec64(A[i],x,P);
      return;
}

LONG VECDOTPROD64s( LONG *A, LONG *B, int n, LONG p ) {
      int i;
      ULONG z[2]; z[0] = z[1] = 0;
      for( i=0; i<n; i++ ) { zfma(z,A[i],B[i]); if( z[1] >= p ) z[1] -= p; }
      zmod(z,p);
      return z[0];
}

void VECPOWMOD64s( LONG *A, LONG m, int n, LONG p ) {
      int i;
      recint P = recip1(p);
      for( i=0; i<n; i++ ) A[i] = POWMOD64s(A[i],m,P);
      return;
}

void VECPOWERS64s( LONG alpha, LONG *A, int n, LONG p ) {
      int i;
      recint P = recip1(p);
      A[0] = 1;
      for( i=1; i<n; i++ ) A[i] = mulrec64(alpha,A[i-1],P);
      return;
}
